import asyncio
from dotenv import load_dotenv
# from langchain_google_genai import ChatGoogleGenerativeAI
import pyautogui
import base64
from browser_use import Agent, Controller, Browser, BrowserConfig, ChatAzureOpenAI, BrowserSession
import time
import numpy as np
import pytesseract
import cv2
import keyboard
from openai import AzureOpenAI
import os
from typing import List
from pydantic import BaseModel
import json 
from flask import Flask, request, jsonify
from pydantic import SecretStr
app = Flask(__name__)

load_dotenv()

endpoint = ''
deployment = ''
subscription_key = ''
api_version = ''
browser_binary_path = ''
tesserect_path = ''

# Initialize Azure OpenAI client with key-based authentication
client = AzureOpenAI(
    azure_endpoint=endpoint,
    api_key=subscription_key,
    api_version=api_version,
)

dctButtonsPosOccr = {}
action_log = {}

def encode_image(path):
    with open(path, "rb") as img_file:
        return base64.b64encode(img_file.read()).decode("utf-8")

def crop_image(img1_path: str, img2_path: str)-> None:
    img1 = cv2.imread(img1_path)
    img2 = cv2.imread(img2_path)

    gray1 = cv2.cvtColor(img1, cv2.COLOR_BGR2GRAY)
    gray2 = cv2.cvtColor(img2, cv2.COLOR_BGR2GRAY)

    diff = cv2.absdiff(gray1, gray2)
    heig,widt = diff.shape
    
    _, thresh = cv2.threshold(diff, 30, 255, cv2.THRESH_BINARY)

    kernel = np.ones((5, 5), np.uint8)
    dilated = cv2.dilate(thresh, kernel, iterations=2)

    contours, _ = cv2.findContours(dilated, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    for cnt in contours:
        x, y, w, h = cv2.boundingRect(cnt)
        cropped_image = img2[y-heig//5:y + h+heig//5, x-widt//5:x + w + widt//5]
        cv2.imwrite(img2_path+"cropped.png", cropped_image)
        return img2_path+"cropped.png"


def crop_images(images: List[str])-> List[str]:
    cropped_images = []
    cropped_images.append(crop_image(images[1], images[0]))
    for image_path in range(1, len(images)):
        cropped_images.append(crop_image(images[image_path-1], images[image_path]))
    return cropped_images


def b64_to_png(b64_string: str, output_file):
    with open(output_file, "wb") as f:
        f.write(base64.b64decode(b64_string))

class ImageComparisonRequest(BaseModel):
    image_paths: List[str]
    comparison_text: str

controller = Controller()
pytesseract.pytesseract.tesseract_cmd = tesserect_path
@controller.action("click on any button using its text or buttons in sequence")
async def click_given_its_textSequence(buttonText:List[str],occurrence:List[int]) -> str:
    global dctButtonsPosOccr
    global action_log

    print(buttonText)
    print(occurrence)
    for i in  range(len(buttonText)):
        button = buttonText[i].split(" ")[0]
        website_screenshot = pyautogui.screenshot()
        screenshot_np = np.array(website_screenshot)
        h,w,_ = screenshot_np.shape
        y1, y2 = 0, int(0.8 * h)
        x1, x2 = 0, w
        screenshot_np = screenshot_np[y1:y2, x1:x2]
        image = cv2.cvtColor(screenshot_np, cv2.COLOR_RGB2BGR)
        cv2.imwrite(f"temp_{button}_{occurrence[i]}.png", image) 
        image = cv2.threshold(image, 180, 255, cv2.THRESH_BINARY)[1]
        match_count = 0
        # print(f"Looking for button: {button} with occurrence: {occurrence[i]}")
        if( button in dctButtonsPosOccr and dctButtonsPosOccr[button][0] == occurrence[i]):
            x, y = dctButtonsPosOccr[button][1], dctButtonsPosOccr[button][2]
            print('\n')
            print(f"Coordinates of '{button}': ({x}, {y})")
            action_log[button] = True
            pyautogui.moveTo(x=x, y=y)
            time.sleep(1)
            pyautogui.click()
            time.sleep(1)
            continue
        results = pytesseract.image_to_data(image, output_type=pytesseract.Output.DICT)
        for j in range(len(results['text'])):
            clicked = False
            # print(len_buttonText,results['text'][i].strip().lower(),buttonText.lower())
            # print(results['text'][j].strip().lower(),button.lower(),match_count,occurrence[i])
            if results['text'][j].strip().lower() == button.lower():
                match_count += 1
                # print(results['text'][j].strip().lower(),button.lower(),match_count,occurrence[i])
                if match_count != occurrence[i]:
                    continue
                x = results['left'][j] + results['width'][j] // 2
                y = results['top'][j] + results['height'][j] // 2
                print('\n')
                print(f"Coordinates of '{button}': ({x}, {y})")
                pyautogui.moveTo(x=x, y=y)
                time.sleep(1)
                pyautogui.click()
                time.sleep(1)
                action_log[f"{button}_{occurrence[i]}"] = True
                dctButtonsPosOccr[button] = [occurrence[i], x, y]
                clicked = True
                break
        if not clicked:
            print("Status: Failed button not found")
            return "Status: Failed button not found"
    print("Status: Success clicked on button")
    return "Status: Success clicked on button"


@controller.action("hover on any button using its text")
async def hover_given_its_text(buttonText:str,occurrence:int = 1) -> None:
    global action_log
    buttonText = buttonText.split(" ")[0]
    website_screenshot = pyautogui.screenshot()
    screenshot_np = np.array(website_screenshot)
    image = cv2.cvtColor(screenshot_np, cv2.COLOR_RGB2BGR)
    cv2.imwrite(f"temp_{buttonText}.png", image) 
    image = cv2.threshold(image, 180, 255, cv2.THRESH_BINARY)[1]
    match_count = 0
    results = pytesseract.image_to_data(image, output_type=pytesseract.Output.DICT)
    for i in range(len(results['text'])):
        # print(len_buttonText,results['text'][i].strip().lower(),buttonText.lower())
        if results['text'][i].strip().lower() == buttonText.lower():
            match_count += 1
            if match_count != occurrence:
                continue
            x = results['left'][i] + results['width'][i] // 2
            y = results['top'][i] + results['height'][i] // 2
            print('\n')
            print(f"Coordinates of '{buttonText}': ({x}, {y})")
            pyautogui.moveTo(x=x, y=y)
            time.sleep(1.5)
            action_log[buttonText] = True
            return 


@controller.action("press shortcut keys given their frequency to press")
async def press_shortcut_key(shortcut_key: str, freq: int) -> str:
    global action_log
    # Split the shortcut key into individual keys
    print('\n')
    keys = shortcut_key.lower().split('+')
    for i in range(freq):
        for key in keys:
            # Press each key
            if(key.strip() == "control"):
                key = "ctrl"
            if(key.strip() == "arrowright"):
                key = "right"
            print(keys)
            keyboard.press(key.strip())
            print("pressed", key.strip())
            time.sleep(0.1)
        time.sleep(0.5)
        for key in reversed(keys):
            # Release each key
            if(key.strip() == "control"):
                key = "ctrl"
            keyboard.release(key.strip())
            print("released", key.strip())
            time.sleep(0.1)
        time.sleep(0.5)
    action_log[shortcut_key] = True
    return "Status: Success pressed shortcut key"


@controller.action("press shortcut keys given their frequency to press and take a screenshot after each press")
async def press_shortcut_key_and_take_screenshot(shortcut_key: str, screenshot_names:List[str] ,freq: int) -> str:
    global action_log
    # Split the shortcut key into individual keys
    print('\n')
    keys = shortcut_key.lower().split('+')
    for i in range(freq):
        for key in keys:
            # Press each key
            if(key.strip() == "control"):
                key = "ctrl"
            if(key.strip() == "arrowright"):
                key = "right"
            print(keys)
            keyboard.press(key.strip())
            print("pressed", key.strip())
            time.sleep(0.1)
        time.sleep(0.2)
        for key in reversed(keys):
            # Release each key
            if(key.strip() == "control"):
                key = "ctrl"
            keyboard.release(key.strip())
            print("released", key.strip())
            time.sleep(0.1)
        time.sleep(1)
        # Take a screenshot after each press
        screenshot = pyautogui.screenshot()
        screen_width, screen_height = pyautogui.size()
        image_np = np.array(screenshot)
        image_np = image_np[:int(0.9*screen_height),:]
        # Save the screenshot with the corresponding name
        cv2.imwrite(screenshot_names[i], cv2.cvtColor(image_np, cv2.COLOR_RGB2BGR))
        print(f"Screenshot saved as {screenshot_names[i]}")
    action_log[shortcut_key] = True
    print("Status: Success pressed shortcut key")
    return "Status: Success pressed shortcut key"

@controller.action("check the images given their paths and text for what to check",param_model=ImageComparisonRequest)
async def check_images(params: ImageComparisonRequest) -> str:
    print(params)
    global action_log
    try:
        new_paths = crop_images(params.image_paths)
    except Exception as e:
        print(f"Error cropping images: {e}")
        new_paths = params.image_paths

    encoded_images = [encode_image(path) for path in new_paths]

    content = [
        {"type": "text", "text": "Please can you verify that " + params.comparison_text}
    ]
    for i, encoded_image in enumerate(encoded_images):
        content.append({
            "type": "image_url",
            "image_url": {
                "url": f"data:image/png;base64,{encoded_image}"
            }
        })
    
    messages = [
        {
            "role": "system",
            "content": [
                {"type": "text", "text": "You are a visual assistant that helps to validate the accessbility bugs. As due to no interactive elements in powerpoint online, we can't use accessibility checkers for verification. So need visual inspection like contrast issues, multiple images insequence for tab ordering, tooltip etc. tell accrding to MAS. Its difficult and we do need to automate so we are using GPT."}
            ]
        },
        { 
            "role": "user",
            "content": content
        }
    ]
    response = client.chat.completions.create(
        model=deployment,
        messages=messages,
        # max_tokens=800,
    )

    # Output
    print(response.choices[0].message.content)
    action_log[str(params)] = True
    return str(response.choices[0].message.content)


@controller.action("Get the location of icons and hover on them")
async def hover_on_icon(icon_name:str)-> str:
    global action_log
    print(icon_name)
    print('\n')
    # Take screenshot
    if icon_name.strip().lower() == "cross" or icon_name.strip().lower() == "close":
        template_path = "./icons/close.png"
    elif icon_name.strip().lower() == "delete":
        template_path = "./icons/delete.png"
    elif icon_name.strip().lower() == "SmartArt":
        template_path = "./icons/smartart.png"
    elif icon_name.strip().lower() == "footer":
        template_path = "./icons/footer.png"
    else:
        print("Not recognized icon name.")
        return "Status: Failed no icon found"
    print(f"Looking for icon: {icon_name} using template: {template_path}\n")
    screenshot = pyautogui.screenshot()
    screenshot_np = np.array(screenshot)

    # Load template icon
    template = cv2.imread(template_path)
    h, w = template.shape[:2]
    print(f"Looking for template path: {template_path}\n")

    # Match template
    result = cv2.matchTemplate(screenshot_np, template, cv2.TM_CCOEFF_NORMED)
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)

    if max_val >= 0.7:  # Adjust threshold as needed
        center_x = max_loc[0] + w // 2
        center_y = max_loc[1] + h // 2
        print(f"Icon found at ({center_x}, {center_y}) with match confidence {max_val}")
        pyautogui.moveTo(center_x, center_y)
        action_log[icon_name] = True
        return  "Status: Success hovered on icon"
    else:
        print("Icon not found.")
        return  "Status: Failed icon not found"


@controller.action("Get the location of icons and click on them")
async def click_on_icon(icon_name:str):
    global action_log
    print(icon_name)
    print('\n')
    # Take screenshot
    if icon_name.strip().lower() == "cross" or icon_name.strip().lower() == "close":
        template_path = "./icons/close.png"
    elif icon_name.strip().lower() == "delete":
        template_path = "./icons/delete.png"
    elif icon_name.strip().lower() == "SmartArt":
        template_path = "./icons/smartart.png"
    elif icon_name.strip().lower() == "footer":
        template_path = "./icons/footer.png"
    else:
        print("Not recognized icon name.")
        return "Status: Failed no icon found"
    screenshot = pyautogui.screenshot()
    screenshot_np = np.array(screenshot)

    # Load template icon
    template = cv2.imread(template_path)
    h, w = template.shape[:2]

    # Match template
    result = cv2.matchTemplate(screenshot_np, template, cv2.TM_CCOEFF_NORMED)
    min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(result)

    if max_val >= 0.7:  # Adjust threshold as needed
        center_x = max_loc[0] + w // 2
        center_y = max_loc[1] + h // 2
        print(f"Icon found at ({center_x}, {center_y}) with match confidence {max_val}")
        pyautogui.moveTo(x=center_x, y=center_y)
        time.sleep(1)
        pyautogui.click()
        time.sleep(1)
        action_log[icon_name] = True
        return 
    else:
        print("Icon not found.")
        return


@controller.action("take screenshot of the page")
async def take_screenshot_of_page(filename: str = "screenshot.png") -> str:
    try:
        global action_log
        screen_width, screen_height = pyautogui.size()
        image = pyautogui.screenshot()
        image_np = np.array(image)
        image_np = image_np[:int(0.9*screen_height),:]
        cv2.imwrite(filename, cv2.cvtColor(image_np, cv2.COLOR_RGB2BGR))  # Save the screenshot
        # image.save(filename) 
        action_log[filename] = True
        print("Status: Screenshot taken and saved as " + filename)
        return "Status: Screenshot taken and saved as " + filename
    except Exception as e:
        return "Status: Failed to take screenshot"


@controller.action("Click on the middle of the screen")
async def click_on_middle_of_screen() -> str:
    global action_log
    try:
        screen_width, screen_height = pyautogui.size()
        middle_x = screen_width // 2
        middle_y = screen_height // 2
        pyautogui.moveTo(middle_x, middle_y)
        time.sleep(1)
        pyautogui.click()
        time.sleep(1)
        action_log["middle_click"] = True
        print("Status: Clicked on the middle of the screen")
        return "Status: Clicked on the middle of the screen"
    except Exception as e:
        return "Status: Failed to click on the middle of the screen"

async def record_activity(agent_obj: Agent):
        website_screenshot = await agent_obj.browser_context.take_screenshot(full_page=True)
        b64_to_png(website_screenshot, "website_screenshot.png")
        return 

async def _async_run_validator(task: str, bug: str):
    model = ChatAzureOpenAI(model=deployment,api_version=api_version,azure_endpoint=endpoint,api_key=subscription_key)
    # model = client
    task = bug  + task
    task = task  +  "\n only return the output of the image checker to results as it will contain details about bug"
    browser = Browser(
      config=BrowserConfig(
            browser_binary_path=browser_binary_path,
            headless=False,
            keep_alive=False,
            # color_scheme="dark",
            # contrast="more",
            # forced_colors="active",
            # args=[
            #     "--force-color-profile=srgb",
            #     "--force-dark-mode",
            #     "--enable-features=WebContentsForceDarkMode",
            #     "--force-prefers-color-scheme=dark",
            #     "--high-contrast"
            # ],
            
            )
        )

    
    agent = Agent(
		task=task,
		llm=model,
        browser=browser,
        controller=controller,
        use_vision=True,
	)

    results = await agent.run(max_steps=10)
    results = results.final_result() 
    await browser.close()
    return json.dumps({"message":results})



import asyncio
def run_validator(task: str, bug: str, endpoint_: str, deployment_: str, subscription_key_: str, api_version_: str, browser_binary_path_: str, tesserect_path_: str) -> str:
    global endpoint, deployment, subscription_key, api_version, browser_binary_path, tesserect_path
    endpoint = endpoint_
    deployment = deployment_
    subscription_key = subscription_key_
    api_version = api_version_
    browser_binary_path = browser_binary_path_
    tesserect_path = tesserect_path_
    return asyncio.run(_async_run_validator(task, bug))
